#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import urllib.parse
from typing import Optional

from pyspark.sql import Column
from pyspark.sql.functions import lit, struct

SNOMED_URI = "http://snomed.info/sct"


def to_coding(
    coding_column: Column, system: str, version: Optional[str] = None
) -> Column:
    """
    Converts a Column containing codes into a Column that contains a Coding struct. The Coding
    struct Column can be used as an input to terminology functions such as `member_of` and
    `translate`.

    :param coding_column: the Column containing the codes
    :param system: the URI of the system the codes belong to
    :param version: the version of the code system
    :return: a Column containing a Coding struct
    """
    id_column = lit(None).alias("id")
    system_column = lit(system).alias("system")
    version_column = lit(version).alias("version")
    display_column = lit(None).alias("display")
    user_selected_column = lit(None).alias("userSelected")
    return struct(
        id_column,
        system_column,
        version_column,
        coding_column.alias("code"),
        display_column,
        user_selected_column,
    )


def to_snomed_coding(coding_column: Column, version: Optional[str] = None) -> Column:
    """
    Converts a Column containing codes into a Column that contains a SNOMED Coding struct. The
    Coding struct Column can be used as an input to terminology functions such as `member_of` and
    `translate`.

    :param coding_column: the Column containing the codes
    :param version: the version of the code system
    :return: a Column containing a Coding struct
    """
    return to_coding(coding_column, SNOMED_URI, version)


def to_ecl_value_set(ecl: str) -> str:
    """
    Converts a SNOMED CT ECL expression into a FHIR ValueSet URI. Can be used with the `member_of`
    function.

    :param ecl: the ECL expression
    :return: the ValueSet URI
    """
    return SNOMED_URI + "?fhir_vs=ecl/" + urllib.parse.quote(ecl, safe="()*!'")



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Optional, Union, Collection

from py4j.java_gateway import JavaObject
from pyspark import SparkContext
from pyspark.sql.column import Column, _to_java_column
from pyspark.sql.functions import lit

from pathling.coding import Coding

CodingArg = Union[Column, str, Coding]
EquivalenceArg = Union[str, Collection[str]]


def _coding_to_java_column(coding: Optional[CodingArg]) -> JavaObject:
    if coding is None:
        return _to_java_column(lit(None))
    else:
        return _to_java_column(
            coding.to_literal() if isinstance(coding, Coding) else coding
        )


def _ensure_collection(
    collection_or_value: Optional[Union[Any, Collection[Any]]]
) -> Optional[Collection[Any]]:
    return (
        collection_or_value
        if isinstance(collection_or_value, Collection)
        and not isinstance(collection_or_value, str)
        else [collection_or_value]
        if collection_or_value is not None
        else None
    )


def _get_jvm() -> JavaObject:
    """
    Gets the py4j JVM associated with this process.
    """
    sc = SparkContext._active_spark_context
    assert sc is not None
    assert sc._jvm is not None
    return sc._jvm


def _invoke_udf(name: str, *args: Any) -> Column:
    """
    Invokes a Terminology UDF function identified by name with args
    and wraps the result with :class:`~pyspark.sql.Column`.
    """
    jf = getattr(_get_jvm().au.csiro.pathling.sql.Terminology, name)
    return Column(jf(*args))


def _invoke_support_function(name: str, *args: Any) -> Any:
    """
    Invokes a Terminology supporting function identified by name with args
    """
    jf = getattr(_get_jvm().au.csiro.pathling.sql.TerminologySupport, name)
    return jf(*args)


class PropertyType:
    """
    Allowed property types.
    """

    STRING = "string"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    DECIMAL = "decimal"
    DATETIME = "dateTime"
    CODE = "code"
    CODING = "Coding"


class Equivalence:
    """
    Concept map equivalences, see https://www.hl7.org/fhir/R4/valueset-concept-map-equivalence.html.
    """

    RELATEDTO = "relatedto"
    EQUIVALENT = "equivalent"
    EQUAL = "equal"
    WIDER = "wider"
    SUBSUMES = "subsumes"
    NARROWER = "narrower"
    SPECIALIZES = "specializes"
    INEXACT = "inexact"
    UNMATCHED = "unmatched"
    DISJOINT = "disjoint"


def member_of(coding: CodingArg, value_set_uri: str) -> Column:
    """
    Takes a Coding or array of Codings column as its input. Returns the column which contains a
    Boolean value, indicating whether any of the input Codings is the member of the specified FHIR
    ValueSet.

    :param coding: a Column containing a struct representation of a Coding or an array of such
           structs.
    :param value_set_uri: an identifier for a FHIR ValueSet
    :return: a Column containing the result of the operation.
    """
    return _invoke_udf("member_of", _coding_to_java_column(coding), value_set_uri)


def translate(
    coding: CodingArg,
    concept_map_uri: str,
    reverse: bool = False,
    equivalences: Optional[EquivalenceArg] = None,
    target: Optional[str] = None,
) -> Column:
    """
    Takes a Coding column as input.  Returns the Column which contains an array of
    Coding value with translation targets from the specified FHIR ConceptMap. There
    may be more than one target concept for each input concept. Only the translation with
    the specified equivalences are returned.
    See also :class:`Equivalence`.
    :param coding: a Column containing a struct representation of a Coding
    :param concept_map_uri: an identifier for a FHIR ConceptMap
    :param reverse: the direction to traverse the map - false results in "source to target"
           mappings, while true results in "target to source"
    :param equivalences: a value of a collection of values from the ConceptMapEquivalence ValueSet
    :param target: identifies the value set in which a translation is sought.  If there's no
           target specified, the server should return all known translations.
    :return: a Column containing the result of the operation (an array of Coding structs).
    """
    return _invoke_udf(
        "translate",
        _coding_to_java_column(coding),
        concept_map_uri,
        reverse,
        _invoke_support_function(
            "equivalenceCodesToEnum", _ensure_collection(equivalences)
        ),
        target,
    )


def subsumes(left_coding: CodingArg, right_coding: CodingArg) -> Column:
    """
    Takes two Coding columns as input. Returns the Column, which contains a
        Boolean value, indicating whether the left Coding subsumes the right Coding.

    :param left_coding: a Column containing a struct representation of a Coding or an array of
           Codings.
    :param right_coding: a Column containing a struct representation of a Coding or an array of
           Codings.
    :return: a Column containing the result of the operation (boolean).
    """
    return _invoke_udf(
        "subsumes",
        _coding_to_java_column(left_coding),
        _coding_to_java_column(right_coding),
    )


def subsumed_by(left_coding: CodingArg, right_coding: CodingArg) -> Column:
    """
    Takes two Coding columns as input. Returns the Column, which contains a
    Boolean value, indicating whether the left Coding is subsumed by the right Coding.

    :param left_coding: a Column containing a struct representation of a Coding or an array of
           Codings.
    :param right_coding: a Column containing a struct representation of a Coding or an array of
           Codings.
    :return: a Column containing the result of the operation (boolean).
    """
    return _invoke_udf(
        "subsumed_by",
        _coding_to_java_column(left_coding),
        _coding_to_java_column(right_coding),
    )


def display(coding: CodingArg) -> Column:
    """
    Takes a Coding column as its input. Returns the Column, which contains the canonical display
    name associated with the given code.

    :param coding: a Column containing a struct representation of a Coding.
    :return: a Column containing the result of the operation (String).
    """
    return _invoke_udf("display", _coding_to_java_column(coding))


def property_of(
    coding: CodingArg, property_code: str, property_type: str = PropertyType.STRING
) -> Column:
    """
    Takes a Coding column as its input. Returns the Column, which contains the values of properties
    for this coding with specified names and types. The type of the result column depends on the
    types of the properties. Primitive FHIR types are mapped to their corresponding SQL primitives.
    Complex types are mapped to their corresponding structs. The allowed property types are: code |
    Coding | string | integer | boolean | dateTime | decimal.
    See also :class:`PropertyType`.

    :param coding: a Column containing a struct representation of a Coding
    :param property_code: the code of the property to retrieve.
    :param property_type: the type of the property to retrieve.
    :return: the Column containing the result of the operation (array of property values)
    """
    return _invoke_udf(
        "property_of", _coding_to_java_column(coding), property_code, property_type
    )


def designation(
    coding: CodingArg, use: Optional[CodingArg] = None, language: Optional[str] = None
) -> Column:
    """
    Takes a Coding column as its input. Returns the Column, which contains the values of
    designations (strings) for this coding for the specified use and language. If the language is
    not provided (is null) then all designations with the specified type are returned regardless of
    their language.

    :param coding: a Column containing a struct representation of a Coding
    :param use: the code with the use of the designations
    :param language: the language of the designations
    :return: the Column containing the result of the operation (array of strings with designation
             values)
    """
    return _invoke_udf(
        "designation",
        _coding_to_java_column(coding),
        _coding_to_java_column(use),
        language,
    )



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from pathling import PathlingContext, Coding, to_coding, subsumes

HERE = os.path.abspath(os.path.dirname(__file__))

pc = PathlingContext.create()

csv = pc.spark.read.options(header=True).csv(
    f'file://{os.path.join(HERE, "data/csv/conditions.csv")}'
)
first_3 = csv.limit(3)
cross_join = first_3.selectExpr(
    "CODE as LEFT", "DESCRIPTION as LEFT_DESCRIPTION"
).crossJoin(first_3.selectExpr("CODE as RIGHT", "DESCRIPTION as RIGHT_DESCRIPTION"))

result_1 = cross_join.withColumn(
    "SUBSUMES",
    subsumes(
        to_coding(cross_join.LEFT, "http://snomed.info/sct"),
        to_coding(cross_join.RIGHT, "http://snomed.info/sct"),
    ),
)
result_2 = result_1.withColumn(
    "LEFT_IS_ENT",
    subsumes(
        # 232208008 |Ear, nose and throat disorder|
        Coding("http://snomed.info/sct", "232208008"),
        to_coding(cross_join.LEFT, "http://snomed.info/sct"),
    ),
)

result_2.select(
    "LEFT", "RIGHT", "LEFT_DESCRIPTION", "RIGHT_DESCRIPTION", "SUBSUMES", "LEFT_IS_ENT"
).show()



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from pathling import PathlingContext, to_snomed_coding, to_ecl_value_set, member_of

HERE = os.path.abspath(os.path.dirname(__file__))

pc = PathlingContext.create()

csv = pc.spark.read.options(header=True).csv(
    f'file://{os.path.join(HERE, "data/csv/conditions.csv")}'
)

VIRAL_INFECTION_ECL = """
    << 64572001|Disease| : (
      << 370135005|Pathological process| = << 441862004|Infectious process|,
      << 246075003|Causative agent| = << 49872002|Virus|
    )
"""

csv.select(
    "CODE",
    "DESCRIPTION",
    member_of(to_snomed_coding(csv.CODE), to_ecl_value_set(VIRAL_INFECTION_ECL)).alias(
        "VIRAL_INFECTION"
    ),
).show()



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional, Sequence

from deprecated import deprecated
# noinspection PyPackageRequirements
from py4j.java_gateway import JavaObject
from pyspark.sql import DataFrame, SparkSession, Column

from pathling.coding import Coding
from pathling.etc import find_jar
from pathling.fhir import MimeType

__all__ = ["PathlingContext"]

EQ_EQUIVALENT = "equivalent"


class StorageType:
  MEMORY: str = "memory"
  DISK: str = "disk"


# noinspection PyProtectedMember
class PathlingContext:
  """
  Main entry point for Pathling API functionality.
  Should be instantiated with the :func:`PathlingContext.create` class method.

  Example use::

      pc = PathlingContext.create(spark)
      patient_df = pc.encode(spark.read.text('ndjson_resources'), 'Patient')
      patient_df.show()
  """

  @property
  def spark(self) -> SparkSession:
    """
    Returns the SparkSession associated with this context.
    """
    return self._spark

  @classmethod
  def create(
      cls,
      spark: Optional[SparkSession] = None,
      max_nesting_level: Optional[int] = 3,
      enable_extensions: Optional[bool] = False,
      enabled_open_types: Optional[Sequence[str]] = (
          "boolean",
          "code",
          "date",
          "dateTime",
          "decimal",
          "integer",
          "string",
          "Coding",
          "CodeableConcept",
          "Address",
          "Identifier",
          "Reference",
      ),
      enable_terminology: Optional[bool] = True,
      terminology_server_url: Optional[str] = "https://tx.ontoserver.csiro.au/fhir",
      terminology_verbose_request_logging: Optional[bool] = False,
      terminology_socket_timeout: Optional[int] = 60_000,
      max_connections_total: Optional[int] = 32,
      max_connections_per_route: Optional[int] = 16,
      terminology_retry_enabled: Optional[bool] = True,
      terminology_retry_count: Optional[int] = 2,
      enable_cache: Optional[bool] = True,
      cache_max_entries: Optional[int] = 200_000,
      cache_storage_type: Optional[str] = StorageType.MEMORY,
      cache_storage_path: Optional[str] = None,
      cache_default_expiry: Optional[int] = 600,
      cache_override_expiry: Optional[int] = None,
      token_endpoint: Optional[str] = None,
      enable_auth: Optional[bool] = False,
      client_id: Optional[str] = None,
      client_secret: Optional[str] = None,
      scope: Optional[str] = None,
      token_expiry_tolerance: Optional[int] = 120,
  ) -> "PathlingContext":
    """
    Creates a :class:`PathlingContext` with the given configuration options. This should only
    be done once within a SparkSession - subsequent calls with different configuration may
    produce an error.

    If no SparkSession is provided, and there is not one already present in this process - a
    new SparkSession will be created.

    If a SparkSession is not provided, and one is already running within the current process,
    it will be reused - and it is assumed that the Pathling library API JAR is already on the
    classpath. If you are running your own cluster, make sure it is on the list of packages.

    If a SparkSession is provided, it needs to include the Pathling library API JAR on its
    classpath. You can get the path for the JAR (which is bundled with the Python package)
    using the `pathling.etc.find_jar` method.

    :param spark: a pre-configured :class:`SparkSession` instance, use this if you need to
           control the way that the session is set up
    :param max_nesting_level: controls the maximum depth of nested element data that is encoded
           upon import. This affects certain elements within FHIR resources that contain
           recursive references, e.g. `QuestionnaireResponse.item
           <https://hl7.org/fhir/R4/questionnaireresponse.html>`_.
    :param enable_extensions: enables support for FHIR extensions
    :param enabled_open_types: the list of types that are encoded within open types,
           such as extensions. This default list was taken from the data types that are common
           to extensions found in widely-used IGs, such as the US and AU base profiles. In
           general, you will get the best query performance by encoding your data with the
           shortest possible list.
    :param enable_terminology: enables the use of terminology functions
    :param terminology_server_url: the endpoint of a FHIR terminology service (R4) that the
           server can use to resolve terminology queries. The default server is suitable for
           testing purposes only.
    :param terminology_verbose_request_logging: setting this option to `True` will enable
           additional logging of the details of requests to the terminology service. Note that
           logging is subject to the Spark logging level, which you can set using
           `SparkContext.setLogLevel`. Verbose request logging is sent to the `DEBUG` logging
           level.
    :param terminology_socket_timeout: the maximum period (in milliseconds) that the server
           should wait for incoming data from the HTTP service
    :param max_connections_total: the maximum total number of connections for the client
    :param max_connections_per_route: the maximum number of connections per route for the client
    :param terminology_retry_enabled: controls whether terminology requests that fail for
           possibly transient reasons (network connections, DNS problems) should be retried
    :param terminology_retry_count: the number of times to retry failed terminology requests
    :param enable_cache: set this to false to disable caching of terminology requests (not
           recommended)
    :param cache_max_entries: sets the maximum number of entries that will be held in memory
    :param cache_storage_type: the type of storage to use for the terminology cache. See
           `StorageType`.
    :param cache_storage_path: the path on disk to use for the cache, required when
           `cache_storage_type` is `disk`
    :param cache_default_expiry: the default expiry time for cache entries (in seconds), used
           when the server does not provide an expiry value
    :param cache_override_expiry: if provided, this value overrides the expiry time provided by
           the terminology server
    :param enable_auth: enables authentication of requests to the terminology server
    :param token_endpoint: an OAuth2 token endpoint for use with the client credentials grant
    :param client_id: a client ID for use with the client credentials grant
    :param client_secret: a client secret for use with the client credentials grant
    :param scope: a scope value for use with the client credentials grant
    :param token_expiry_tolerance: the minimum number of seconds that a token should have
           before expiry when deciding whether to send it with a terminology request
    :return: a :class:`PathlingContext` instance initialized with the specified configuration
    """
    spark = (
        spark
        or SparkSession.getActiveSession()
        or SparkSession.builder.config("spark.jars", find_jar()).getOrCreate()
    )
    jvm = spark._jvm

    # Build an encoders configuration object from the provided parameters.
    encoders_config = (
      jvm.au.csiro.pathling.config.EncodingConfiguration.builder()
      .maxNestingLevel(max_nesting_level)
      .enableExtensions(enable_extensions)
      .openTypes(jvm.java.util.HashSet(enabled_open_types))
      .build()
    )

    # Build a terminology client configuration object from the provided parameters.
    client_config = (
      jvm.au.csiro.pathling.config.HttpClientConfiguration.builder()
      .socketTimeout(terminology_socket_timeout)
      .maxConnectionsTotal(max_connections_total)
      .maxConnectionsPerRoute(max_connections_per_route)
      .retryEnabled(terminology_retry_enabled)
      .retryCount(terminology_retry_count)
      .build()
    )

    # Build a terminology cache configuration object from the provided parameters.
    cache_storage_type_enum = (
      jvm.au.csiro.pathling.config.HttpClientCachingStorageType.fromCode(
        cache_storage_type
      )
    )
    cache_config = (
      jvm.au.csiro.pathling.config.HttpClientCachingConfiguration.builder()
      .enabled(enable_cache)
      .maxEntries(cache_max_entries)
      .storageType(cache_storage_type_enum)
      .storagePath(cache_storage_path)
      .defaultExpiry(cache_default_expiry)
      .overrideExpiry(cache_override_expiry)
      .build()
    )

    # Build a terminology authentication configuration object from the provided parameters.
    auth_config = (
      jvm.au.csiro.pathling.config.TerminologyAuthConfiguration.builder()
      .enabled(enable_auth)
      .tokenEndpoint(token_endpoint)
      .clientId(client_id)
      .clientSecret(client_secret)
      .scope(scope)
      .tokenExpiryTolerance(token_expiry_tolerance)
      .build()
    )

    # Build a terminology configuration object from the provided parameters.
    terminology_config = (
      jvm.au.csiro.pathling.config.TerminologyConfiguration.builder()
      .enabled(enable_terminology)
      .serverUrl(terminology_server_url)
      .verboseLogging(terminology_verbose_request_logging)
      .client(client_config)
      .cache(cache_config)
      .authentication(auth_config)
      .build()
    )

    jpc: JavaObject = jvm.au.csiro.pathling.library.PathlingContext.create(
      spark._jsparkSession, encoders_config, terminology_config
    )
    return PathlingContext(spark, jpc)

  def __init__(self, spark: SparkSession, jpc: JavaObject) -> None:
    self._spark: SparkSession = spark
    self._jpc: JavaObject = jpc

  def _wrap_df(self, jdf: JavaObject) -> DataFrame:
    #
    # Before Spark v3.3 Dataframes were constructs with SQLContext, which was available
    # in `_wrapped` attribute of SparkSession.
    # Since v3.3 Dataframes are constructed with SparkSession instance directly.
    #
    return DataFrame(
      jdf,
      self._spark._wrapped if hasattr(self._spark, "_wrapped") else self._spark,
    )

  def encode(
      self,
      df: DataFrame,
      resource_name: str,
      input_type: Optional[str] = None,
      column: Optional[str] = None,
  ) -> DataFrame:
    """
    Takes a dataframe with a string representations of FHIR resources in the given column and
    encodes the resources of the given types as Spark dataframe.

    :param df: a :class:`DataFrame` containing the resources to encode.
    :param resource_name: the name of the FHIR resource to extract (Condition, Observation,
           etc.)
    :param input_type: the mime type of input string encoding. Defaults to
           `application/fhir+json`.
    :param column: the column in which the resources to encode are stored. If 'None' then the
           input dataframe is assumed to have one column of type string.
    :return: a :class:`DataFrame` containing the given type of resources encoded into Spark
             columns
    """

    return self._wrap_df(
      self._jpc.encode(
        df._jdf, resource_name, input_type or MimeType.FHIR_JSON, column
      )
    )

  def encode_bundle(
      self,
      df: DataFrame,
      resource_name: str,
      input_type: Optional[str] = None,
      column: Optional[str] = None,
  ) -> DataFrame:
    """
    Takes a dataframe with a string representations of FHIR bundles  in the given column and
    encodes the resources of the given types as Spark dataframe.

    :param df: a :class:`DataFrame` containing the bundles with the resources to encode.
    :param resource_name: the name of the FHIR resource to extract (Condition, Observation,
           etc.)
    :param input_type: the MIME type of the input string encoding. Defaults to
           `application/fhir+json`.
    :param column: the column in which the resources to encode are stored. If 'None' then the
           input dataframe is assumed to have one column of type string.
    :return: a :class:`DataFrame` containing the given type of resources encoded into Spark
             columns
    """
    return self._wrap_df(
      self._jpc.encodeBundle(
        df._jdf, resource_name, input_type or MimeType.FHIR_JSON, column
      )
    )

  @deprecated(reason="You should use the 'udfs.member_of' UDF instead")
  def member_of(
      self,
      df: DataFrame,
      coding_column: Column,
      value_set_uri: str,
      output_column_name: str,
  ):
    """
    Takes a dataframe with a Coding column as input. A new column is created which contains a
    Boolean value, indicating whether the input Coding is a member of the specified FHIR
    ValueSet.

    :param df: a DataFrame containing the input data
    :param coding_column: a Column containing a struct representation of a Coding
    :param value_set_uri: an identifier for a FHIR ValueSet
    :param output_column_name: the name of the result column
    :return: A new dataframe with an additional column containing the result of the operation
    """
    return self._wrap_df(
      self._jpc.memberOf(
        df._jdf, coding_column._jc, value_set_uri, output_column_name
      )
    )

  @deprecated(reason="You should use the 'udfs.translate' UDF instead")
  def translate(
      self,
      df: DataFrame,
      coding_column: Column,
      concept_map_uri: str,
      reverse: Optional[bool] = False,
      equivalence: Optional[str] = EQ_EQUIVALENT,
      target: Optional[str] = None,
      output_column_name: Optional[str] = "result",
  ):
    """
    Takes a dataframe with a Coding column as input. A new column is created which contains
    the array of Codings value with translation targets from the specified FHIR ConceptMap.
    There may be more than one target concept for each input concept.

    :param df: a DataFrame containing the input data
    :param coding_column: a Column containing a struct representation of a Coding
    :param concept_map_uri: an identifier for a FHIR ConceptMap
    :param reverse: the direction to traverse the map - false results in "source to target"
           mappings, while true results in "target to source"
    :param equivalence: a comma-delimited set of values from the ConceptMapEquivalence ValueSet
    :param target: identifies the value set in which a translation is sought.  If there is no
           target specified, the server should return all known translations.
    :param output_column_name: the name of the result column
    :return: A new dataframe with an additional column containing the result of the operation.
    """
    return self._wrap_df(
      self._jpc.translate(
        df._jdf,
        coding_column._jc,
        concept_map_uri,
        reverse,
        equivalence,
        target,
        output_column_name,
      )
    )

  @deprecated(reason="You should use the 'udfs.subsumes' UDF instead")
  def subsumes(
      self,
      df: DataFrame,
      output_column_name: str,
      left_coding_column: Optional[Column] = None,
      right_coding_column: Optional[Column] = None,
      left_coding: Optional[Coding] = None,
      right_coding: Optional[Coding] = None,
  ):
    """
    Takes a dataframe with two Coding columns. A new column is created which contains a
    Boolean value, indicating whether the left Coding subsumes the right Coding.

    :param df: a DataFrame containing the input data
    :param left_coding_column: a Column containing a struct representation of a Coding,
           for the left-hand side of the subsumption test
    :param right_coding_column: a Column containing a struct representation of a Coding,
           for the right-hand side of the subsumption test
    :param left_coding: a Coding object for the left-hand side of the subsumption test
    :param right_coding: a Coding object for the right-hand side of the subsumption test
    :param output_column_name: the name of the result column
    :return: A new dataframe with an additional column containing the result of the operation.
    """
    if (left_coding_column is None and left_coding is None) or (
        right_coding_column is None and right_coding is None
    ):
      raise ValueError(
        "Must provide either left_coding_column or left_coding, and either "
        "right_coding_column or right_coding"
      )
    left_column = left_coding.to_literal() if left_coding else left_coding_column
    right_column = (
      right_coding.to_literal() if right_coding else right_coding_column
    )
    return self._wrap_df(
      self._jpc.subsumes(
        df._jdf, left_column._jc, right_column._jc, output_column_name
      )
    )



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from pathling import PathlingContext

HERE = os.path.abspath(os.path.dirname(__file__))

pc = PathlingContext.create()

# Read each line from the NDJSON into a row within a Spark data set.
ndjson_dir = os.path.join(HERE, "data/resources/")
json_resources = pc.spark.read.text(ndjson_dir)

# Convert the data set of strings into a structured FHIR data set.
patients = pc.encode(json_resources, "Patient")

# Do some stuff.
patients.select("id", "gender", "birthDate").show()



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from pathling import PathlingContext, Coding, to_snomed_coding, designation

HERE = os.path.abspath(os.path.dirname(__file__))

pc = PathlingContext.create()

csv = pc.spark.read.options(header=True).csv(
    f'file://{os.path.join(HERE, "data/csv/conditions.csv")}'
)

# Obtain display name for snomed codes

SNOMED_FULLY_SPECIFIED_NAME = Coding.of_snomed("900000000000003001")

result = csv.withColumn(
    "FULL_NAME",
    designation(to_snomed_coding(csv.CODE), SNOMED_FULLY_SPECIFIED_NAME, "en").getItem(
        0
    ),
)
result.select("CODE", "DESCRIPTION", "FULL_NAME").show()



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Optional

from pyspark.sql.functions import lit, struct

from pathling.functions import SNOMED_URI


class Coding:
    """
    A Coding represents a code in a code system.
    See: https://hl7.org/fhir/R4/datatypes.html#Coding
    """

    def __init__(
        self,
        system: str,
        code: str,
        version: Optional[str] = None,
        display: Optional[str] = None,
        user_selected: Optional[bool] = None,
    ):
        """
        :param system: a URI that identifies the code system
        :param code: the code
        :param version: a URI that identifies the version of the code system
        :param display: the display text for the Coding
        :param user_selected: an indicator of whether the Coding was chosen directly by the user
        """
        self.system = system
        self.code = code
        self.version = version
        self.display = display
        self.user_selected = user_selected

    def to_literal(self):
        """
        Converts a Coding into a Column that contains a Coding struct. The Coding
        struct Column can be used as an input to terminology functions such as `member_of` and
        `translate`.

        :return: a Column containing a Coding struct
        """
        id_column = lit(None).alias("id")
        system_column = lit(self.system).alias("system")
        version_column = lit(self.version).alias("version")
        code_column = lit(self.code).alias("code")
        display_column = lit(self.display).alias("display")
        user_selected_column = lit(self.user_selected).alias("userSelected")
        return struct(
            id_column,
            system_column,
            version_column,
            code_column,
            display_column,
            user_selected_column,
        )

    @classmethod
    def of_snomed(
        cls,
        code: str,
        version: Optional[str] = None,
        display: Optional[str] = None,
        user_selected: Optional[bool] = None,
    ) -> "Coding":
        """
        Creates a SNOMED Coding.

        :param code: the code
        :param version: a URI that identifies the version of the code system
        :param display: the display text for the Coding
        :param user_selected: an indicator of whether the Coding was chosen directly by the user
        :return: a SNOMED coding with given arguments.
        """
        return Coding(SNOMED_URI, code, version, display, user_selected)



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .coding import Coding
from .context import PathlingContext, StorageType
from .etc import find_jar
from .fhir import MimeType, Version
from .functions import to_coding, to_snomed_coding, to_ecl_value_set
from .udfs import (
    member_of,
    translate,
    subsumes,
    subsumed_by,
    property_of,
    display,
    designation,
    PropertyType,
    Equivalence,
)

__all__ = [
    "PathlingContext",
    "StorageType",
    "MimeType",
    "Version",
    "Coding",
    "member_of",
    "translate",
    "subsumes",
    "subsumed_by",
    "property_of",
    "display",
    "designation",
    "PropertyType",
    "Equivalence",
    "to_coding",
    "to_snomed_coding",
    "to_ecl_value_set",
    "find_jar",
]



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


class MimeType:
    """
    Constants for FHIR encoding mime types.
    """

    FHIR_JSON: str = "application/fhir+json"
    FHIR_XML: str = "application/fhir+xml"


class Version:
    """
    Constants for FHIR versions.
    """

    R4: str = "R4"



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from pathling import PathlingContext
from pathling.functions import to_coding, to_ecl_value_set

HERE = os.path.abspath(os.path.dirname(__file__))

pc = PathlingContext.create()

csv = pc.spark.read.options(header=True).csv(
    f'file://{os.path.join(HERE, "data/csv/conditions.csv")}'
)

result = pc.member_of(
    csv,
    to_coding(csv.CODE, "http://snomed.info/sct"),
    to_ecl_value_set(
        """
<< 64572001|Disease| : (
  << 370135005|Pathological process| = << 441862004|Infectious process|,
  << 246075003|Causative agent| = << 49872002|Virus|
)
                      """
    ),
    "VIRAL_INFECTION",
)
result.select("CODE", "DESCRIPTION", "VIRAL_INFECTION").show()



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import glob
import os
import sys
from typing import Union

from setuptools import setup, find_packages

HERE = os.path.dirname(__file__)
ROOT_DIR = os.path.abspath(HERE)
VERSION_FILE = os.path.join(HERE, "pathling", "_version.py")

#
# Read __version__ from file
#
__version__: Union[str, None] = None
if os.path.isfile(VERSION_FILE):
    with open(VERSION_FILE) as vf:
        exec(vf.read())
if not __version__:
    print(
        "ERROR: Cannot read __version__ from file '%s'." % VERSION_FILE, file=sys.stderr
    )
    exit(1)

#
# Check for existence of uber-jar for packaging
#
JARS_DIR = os.path.join(HERE, "target", "dependency")
UBER_JAR_GLOB = os.path.join(JARS_DIR, "library-api-*-all.jar")
jar_files = glob.glob(UBER_JAR_GLOB)
if not jar_files:
    print(
        "ERROR: Cannot find library API JAR in: '%s'.\nMaybe try to run 'mvn package' ..."
        % JARS_DIR,
        file=sys.stderr,
    )
    exit(1)
if len(jar_files) > 1:
    print(
        "ERROR: Multiple library API JARs found in: '%s'.\nMaybe try to run 'mvn clean' ..."
        % JARS_DIR,
        file=sys.stderr,
    )
    exit(1)
#
# Read the long description
#
with open("README.md") as f:
    long_description = f.read()

setup(
    name="pathling",
    packages=find_packages()
    + ["pathling.jars"],  # this must be the same as the name above
    version=__version__,
    description="Python API for Pathling",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Australian e-Health Research Centre, CSIRO",
    author_email="ontoserver-support@csiro.au",
    url="https://github.com/aehrc/pathling",
    keywords=["pathling", "fhir", "analytics", "spark", "standards", "terminology"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    license="Apache License, version 2.0",
    python_requires=">=3.7",
    install_requires=["pyspark>=3.1.0", "deprecated>=1.2.13"],
    include_package_data=True,
    package_dir={
        "pathling.jars": "target/dependency",
    },
    package_data={
        "pathling.jars": ["*-all.jar"],
    },
    data_files=[
        ("share/pathling/examples", glob.glob("examples/*.py")),
        (
            "share/pathling/examples/data/resources",
            glob.glob("examples/data/resources/*.ndjson"),
        ),
        (
            "share/pathling/examples/data/bundles",
            glob.glob("examples/data/bundles/*.json"),
        ),
    ],
)



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import glob
import os

import pkg_resources


def find_jar(verbose: bool = False) -> str:
    """
    Gets the path to the Pathling library JAR, bundled with the python distribution.
    """
    # try to find the distribution jars first
    verbose and print("Package name is: %s" % __name__)
    verbose and print("Filename name is: %s" % __file__)

    jars_dir = pkg_resources.resource_filename(__name__, "jars")

    verbose and print("Packaged jars dir: %s" % jars_dir)
    if not os.path.isdir(jars_dir):
        verbose and print(
            "Package resource dir '%s' does not exist. Maybe a development install?"
            % jars_dir
        )
        # then it can be a development install
        project_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), os.pardir)
        )
        verbose and print("Project dir is: %s" % project_dir)
        jars_dir = os.path.join(project_dir, "target", "dependency")
    jar_file = glob.glob(os.path.join(jars_dir, "library-api-*-all.jar"))
    verbose and print("Found jar file(s): %s" % jar_file)
    if not jar_file:
        raise RuntimeError("Pathling jar not present at: %s" % jars_dir)
    return jar_file[0]



#  Copyright 2023 Commonwealth Scientific and Industrial Research
#  Organisation (CSIRO) ABN 41 687 119 230.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from pathling import (
    PathlingContext,
    to_snomed_coding,
    property_of,
    display,
    PropertyType,
)

HERE = os.path.abspath(os.path.dirname(__file__))

pc = PathlingContext.create()

csv = pc.spark.read.options(header=True).csv(
    f'file://{os.path.join(HERE, "data/csv/conditions.csv")}'
)

# Get the parent codes for each code in the dataset.
parents = csv.withColumn(
    "PARENTS",
    property_of(to_snomed_coding(csv.CODE), "parent", PropertyType.CODE),
)
# Split each parent code into a separate row.
exploded_parents = parents.selectExpr(
    "CODE", "DESCRIPTION", "explode_outer(PARENTS) AS PARENT"
)
# Retrieve the preferred term for each parent code.
with_displays = exploded_parents.withColumn(
    "PARENT_DISPLAY", display(to_snomed_coding(exploded_parents.PARENT))
)
with_displays.show()
